/*-------------------------------------------------------------------------
 *  cr_shanon.c:  Code to save/restore shared-anonymous memory mappings
 *
 *  2007, The Regents of the University of California, through Lawrence
 *  Berkeley National Laboratory (subject to receipt of any required
 *  approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 *  Portions may be copyrighted by others, as may be noted in specific
 *  copyright notices within specific files.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_shanon.c,v 1.6 2007/05/04 02:09:47 phargrov Exp $
 *
 * This file handles shared-anonymous memory mappings
 *-----------------------------------------------------------------------*/

#include "cr_module.h"
#include <linux/mman.h>

// Descriptor for shared/anonymous mapping
// Used both in-memory and on-disk
struct cr_shanon_desc {
  /* What is mapped: */
    void *		object; /* The (struct file *) at checkpoint time */
    unsigned long	pgoff;  /* units of PAGE_SIZE */
  /* Where and how is it mapped: */
    unsigned long	start, end;
    unsigned long	flags;
};

/* Read the shared-anonymous mapping metadata from the context file
 *
 * Returns 0 on success, or <0 on error
 */
long
cr_load_shanon_maps(cr_rstrt_proc_req_t *proc_req)
{
    struct cr_shanon_desc *desc_tbl = NULL;
    size_t size = 0;
    int count = 0;
    int retval;

    /* Read size (in bytes) of table */
    retval = cr_kread(proc_req->file, &size, sizeof(size));
    if (retval < sizeof(size)) {
        CR_ERR("short read: %d of %d bytes", retval, (int)sizeof(size));
        retval = -EIO;
        goto err;
    }
    if (!size) goto out_notbl;
    count = size / sizeof(struct cr_shanon_desc);
#if CRI_DEBUG
    if (size != count * sizeof(struct cr_shanon_desc)) {
        CR_ERR("read invalid shanon tbl size: %ld is not a multiple of %ld bytes", (long)size, (long)sizeof(struct cr_shanon_desc));
        retval = -EINVAL;
        goto err;
    }
#endif

    /* Allocate space */
    desc_tbl = vmalloc(size);
    if (!desc_tbl) {
	CR_ERR("Failed to allocate %ld bytes for shanon table", (long)size);
        retval = -ENOMEM;
        goto err;
    }

    /* Read the table */
    retval = cr_kread(proc_req->file, desc_tbl, size);
    if (retval < size) {
        CR_ERR("short read: %d of %d bytes", retval, (int)size);
        retval = -EIO;
        goto err;
    }

    /* Install */
out_notbl:
    proc_req->shanon_cnt = count;
    proc_req->shanon_tbl = desc_tbl;

err:
    return retval;
}

/* Read the shared-anonymous data from the context file
 *
 * Returns 0 on success, or <0 on error
 */
long
cr_load_shanon_data(cr_rstrt_proc_req_t *proc_req)
{
    const int count = proc_req->shanon_cnt;
    const struct cr_shanon_desc *desc = proc_req->shanon_tbl;
    struct mm_struct *mm = current->mm;
    int retval, i;

    retval = 0;
    for (i=0; i<count; ++i, ++desc) {
        struct file *filp = NULL;
	unsigned long flags = MAP_FIXED | MAP_SHARED;
	unsigned long len = desc->end - desc->start;
	unsigned long map_addr;
	int prot;
	    
        (void)cr_find_object(proc_req->req->map, desc->object, (void **)&filp);

        if (desc->flags & VM_DENYWRITE) flags |= MAP_DENYWRITE;
	prot = PROT_READ|PROT_WRITE|PROT_EXEC;

        DOWN_WRITE_MMAP_SEM(mm);
        map_addr = do_mmap(filp, desc->start, len, prot, flags, 0);
        UP_WRITE_MMAP_SEM(mm);
        if (map_addr != desc->start) {
	    CR_ERR("Failed to %smap shared anonymous space", filp ? "re" : "");
	    retval = -EINVAL;
	    goto err;
	}

	/* XXX: Assumes (as does the pipe buffer code) that we are serialized
	 * at a higher level to avoid the lookup-vs-insert race. */
	if (!filp) {
            struct vm_area_struct *map;
            DOWN_READ_MMAP_SEM(mm);
            map = find_vma(mm, desc->start);
	    filp = (map && (map->vm_start == desc->start)) ? map->vm_file : NULL;
            UP_READ_MMAP_SEM(mm);
	    if (!filp) {
	        CR_ERR("Failed to locate newborn shared anonymous space");
	        retval = -EINVAL;
	        goto err;
	    }
            (void)cr_insert_object(proc_req->req->map, desc->object, (void *)filp);
	}

	/* Load (possibly empty) list of pages */
        retval = vmadump_load_page_list(NULL, proc_req->file, (desc->flags & VM_EXEC));
	if (retval) { goto err; }

	/* Set required perms */
	prot = 0;
        if (desc->flags & VM_READ)  prot |= PROT_READ;
        if (desc->flags & VM_WRITE) prot |= PROT_WRITE;
        if (desc->flags & VM_EXEC)  prot |= PROT_EXEC;
	if (sys_mprotect(desc->start, len, prot)) {
	    CR_WARN("Ignoring sys_mprotect() failure.");
        }
    }

err:
    return retval;
}

/* Save the shared-anonymous mappings metadata for the current mm.
 * They are both written to the context file, and saved in the proc_req
 * for later use.
 *
 * Returns bytes written (>=0), or <0 on error
 */
long
cr_save_shanon_maps(cr_chkpt_proc_req_t *proc_req)
{
    struct mm_struct *mm = current->mm;
    struct vm_area_struct *map;
    struct cr_shanon_desc *desc_tbl = NULL;
    struct cr_shanon_desc *desc;
    size_t size = 0;
    int count = 0;
    int retval;

    DOWN_READ_MMAP_SEM(mm);
    /* First pass just counts maps ... */
    for (map = mm->mmap; map ; map = map->vm_next) {
	count += ((map->vm_flags & VM_SHARED) && map->vm_file &&
		  vmad_dentry_unlinked(map->vm_file->f_dentry));
    }
    /* ... then we allocate the table ... */
    if (!count) goto out_notbl;
    size = count * sizeof(struct cr_shanon_desc);
    desc_tbl = vmalloc(size);
    if (!desc_tbl) {
	retval = -ENOMEM;
	goto out_notbl;
    }
    /* ... and the second pass fills the table. */
    desc = desc_tbl;
    for (map = mm->mmap; map ; map = map->vm_next) {
	if ((map->vm_flags & VM_SHARED) && map->vm_file &&
	    vmad_dentry_unlinked(map->vm_file->f_dentry)) {
	    desc->object = map->vm_file;
	    desc->pgoff  = map->vm_pgoff;
	    desc->start  = map->vm_start;
	    desc->end    = map->vm_end;
	    desc->flags  = map->vm_flags;
	    ++desc;
	}
    }
out_notbl:
    proc_req->shanon_cnt = count;
    proc_req->shanon_tbl = desc_tbl;
    UP_READ_MMAP_SEM(mm);

    /* write the size in bytes */
    retval = cr_kwrite(proc_req->file, &size, sizeof(size));
    if (retval < sizeof(size)) {
        CR_ERR("short write: %d of %d bytes", retval, (int)sizeof(size));
        retval = -EIO;
        goto err;
    }

    /* write the table (if any) */
    if (size) {
	long w = cr_kwrite(proc_req->file, desc_tbl, size);
        if (w < size) {
            CR_ERR("short write: %ld of %ld bytes", w, (long)size);
            retval = -EIO;
            goto err;
        }
        retval += w;
    }

err:
    return retval;
}

/* Save the shared-anonymous mapped data for the current mm.
 *
 * Returns bytes written (>=0), or <0 on error
 */
long
cr_save_shanon_data(cr_chkpt_proc_req_t *proc_req)
{
    const int count = proc_req->shanon_cnt;
    const struct cr_shanon_desc *desc = proc_req->shanon_tbl;
    long w, retval;
    int i;

    retval = 0;
    for (i = 0; i < count; ++i, ++desc) {
	struct file *filp = (struct file *)desc->object;
	unsigned long start = desc->start;
	unsigned long end = desc->end;
	unsigned long len = end - start;
	loff_t size = cr_i_size_read(filp->f_dentry->d_inode);

        if (cr_insert_object(proc_req->req->map, filp, (void *)1UL)) {
	    /* Somebody dumped this one already */
	    /* Save an empty page list (so we always have something to read back) */
	    start = end = 0;
	} else if (len < size) {
	    /* We warn only once per object */
	    CR_WARN("[%d] May fail to restore incomplete mapping (size=%lu, mapped=%lu)",
			current->pid, (unsigned long)size, len);
	}
 
        w = vmadump_store_page_list(NULL, proc_req->file, start, end);
	if (w < 0) {
	    retval = -EIO;
	    goto err;
	}
        retval += w;
    }

err:
    return retval;
}
